##########################################################################
# Copyright (c) 2023-2024 Accenture GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Collection of auxiliaries in incubation phase
*********************************************

:module: incubation.auxiliaries

.. currentmodule:: incubation.auxiliaries

"""
